// ********************************************************************
// ROOT macro: generate_pp_cs
// Reads EPICS 2017 data contained in EPDL2017.ALL (placed in the same
// directory as this macro) for pair production process(pair and triplet)
// for all chemical elements (Z=1 to 100).
// Generates 100 cross section data files in the directory 
// pair_EPICS2017(one file per element) in compliance with the livermore 
// GEANT4 format.
// The macro creates 4 directories, in which the generated files are placed :
//    pairdata   => intermediate files containing pure pair production cs
//    tripdata   => intermediate files containing triplet production cs before interpolation
//    tripInterp => intermediate files containing the interpolated 
//                                    version of triplet production cs
//    pair       => final files for total pair production = pair + interpolated triplet
// Authors: Zhuxin LI @CENBG,France
// 2019.12.09
// ********************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
//
bool IsEqual(double a,double b , double eps,double releps)
{
	if(a==b)
	{
		return true;
	}
	
	if(fabs(a-b)<=releps*fabs(b))
	{
		return true;
	}
	
	if(fabs(a-b)<eps)
	{
      return true;
	}
	
	
	return false;
}
double eps=1e-15;//absolut difference
double releps=1e-10;//relative difference

void extract_pair_trip_cs()
{
	gROOT->Reset();
    gStyle->SetPalette(1);
    gROOT->SetStyle("Plain");
    gStyle->SetOptStat(000000);
    
	// open original data file EPDL2017.ALL
	// placed in the same directory as this macro
    FILE * fp = fopen("EPDL2017.ALL","r");  
    double e; //energy 
	double cs; // cross section
    Int_t ncols = 0;
   
    char str[64]= "";   
	Int_t z = 1;
    
    while (1)
    {
        // read one line of 63 characters		
        if(  fgets (str, 64, fp)!= NULL ) 
        {
// ********************************************************************
// Read pure pair production for 100 elements
// ********************************************************************

			// starting from this line :
			// "74  0  0  0.0         0.0         0.0         0.0         0.0", 
			// which is typical for pure pair production for one element
            if (!strcmp(str,
                        "74  0  0  0.0         0.0         0.0         0.0         0.0\r\n"))
            {
				
                cout << "************************ Z=" << z << endl; // atomic number
                
				// the directory "pairdata" will be created in the generate_pp_cs method
				// and all pure pair production data files will be placed in this directory
                char filenameFormat[64] = "epics2017/pairdata/pp-pair-cs-%d.dat"; 
				char filename[64];
				sprintf (filename,filenameFormat,z); //appends z to the file name
                
                cout << filename << endl;
				
                FILE * fp2 = fopen(filename,"w");	
				// variable input will contain all energy-cross section table for one element				
                string input = ""; 
				// to calculate the total number of eneryg-cross section points
				Int_t loop = 0; 
				// to keep the minimum and maximum energy of the points for one element
				double startEnergy, endEnergy;  
				while (1)
                {
					char energy_cs [80] = "";  // a temporary variable
					char data[73] = "";					
					// read the data line (energy - coss section) line by line
					// each line is terminated by a \n which interrupts the reading
					fgets (data, 73, fp); 
					//each table line ends by  "           1"   with 1 written at column 72
					if (strcmp(data,
                        "                                                                       1")==0) 
						{
							// we finish reading the energy - cross section data for the element
							// so now we create the two header lines and put them 
							// at the beginning of the table
							cout<<"ending---------"<<data<<endl;
							endEnergy = e;
							char headers [100]; //a temporary variable
							
							//save the two header lines in the variable headers
							sprintf(headers,"%.9e %.9e %d\n%d\n", startEnergy, endEnergy, loop,loop ); 
							input = headers + input; //concatenate two header lines with data table
							// write the complete data for one element in fp2 file
							fprintf(fp2,"%s",input.c_str());							
							break;
						}
                    // variable data contains one line energy-cross section					
                    ncols =  sscanf(data,"%lf %lf",&e,&cs); // read energy and cross section respectively from the 
															// valable data keeping the same significant numbers
					// change unit of cross section from barn to mm2
					cs = cs*1e-22; 
					
					// save energy-cross section point in the variable energy_cs
					sprintf(energy_cs,"%.9e %.9e\n",e,cs); 
					
					// sprintf(energy_cs,"%.9E %.9E\n",e,cs);
					if (loop == 0) 
					{
						startEnergy = e;
					}				
					// concatenate energy-cross section one by one
					input = input + energy_cs;  
					// counts the number of data points (lines)
					loop++; 
                    
                }
                fclose(fp2);
                //z++;
                
               
            } 
			// starting from this line :
			// "75  0  0  0.0         0.0         0.0         0.0         0.0", 
			// which is typical for triplet production for one element
// ********************************************************************
// Read triplet production for 100 elements
// ********************************************************************
			else if (!strcmp(str,
                        "75  0  0  0.0         0.0         0.0         0.0         0.0\r\n"))
            {
                cout << "************************ Z=" << z << endl; // atomic number
				
                // the directory "tripdata" will be created in the generate_pp_cs method
				// and all triplet production data files will be placed in this directory
				char filenameFormat[64] = "epics2017/tripdata/pp-trip-cs-%d.dat";  
				char filename[64];
				sprintf (filename,filenameFormat,z); //appends z to the file name
                
                cout << filename << endl;
                FILE * fp3 = fopen(filename,"w");
				
				// variable input will contain all energy-cross section table for one element
                string input; 
				
               	// to calculate the total number of eneryg-cross section points		
				Int_t loop = 0; 
				// to keep the minimum and maximum energy of the points for one element
				double startEnergy, endEnergy;// to keep the minimum and maximum energy of the points for one element
				
				while (1)
                {
					char energy_cs [80];  // a temporary variable
					char data[73];
					// read the data line (energy - coss section) line by line
					// each line is terminated by a \n which interrupts the reading
					fgets (data, 73, fp);   // read the data line (energy - coss section) line by line
					
					//each table line ends by  "           1"   with 1 written at column 72
					if (strcmp(data,
                        "                                                                       1")==0)
						{
							// we finish reading the energy - cross section data for the element
							// so now we create the two header lines and put them 
							// at the beginning of the table
							cout<<"ending---------"<<data<<endl;
							endEnergy = e;
							char headers [100]; // a temporary variable
							//save the two header lines in the variable headers
							sprintf(headers,"%.9e %.9e %d\n%d\n", startEnergy, endEnergy, loop,loop );  //save the two header lines in the variable headers
							input = headers + input; //concatenate two header lines with data table
							// write the complete data for one element in fp3 file
							fprintf(fp3,"%s",input.c_str());							
							break;
						}
                    // variable data contains one line energy-cross section
                    ncols =  sscanf(data,"%lf %lf",&e,&cs); // read energy and cross section respectively from the 
															// valable data keeping the same significant numbers
					// change unit of cross section from barn to mm2
					cs=cs*1e-22; 
					// save energy-cross section point in the variable energy_cs
					sprintf(energy_cs,"%.9e %.9e\n",e,cs); 
					
					//sprintf(energy_cs,"%.9E %.9E\n",e,cs);
					if (loop == 0) 
					{
						startEnergy = e;
					}				
					// concatenate energy-cross section one by one
					input = input + energy_cs; 
					// counts the number of data points (lines)
				    loop++;
                    
                }
                
                fclose(fp3);
                z++; //switch to the next element
			}
		}
        else break;
    }

    fclose(fp);    
	
}

void create_pp_cs()

// ********************************************************************
	// we interpolate triplet data so that we get the same energy points 
	// for pair (taken as a reference) and triplet 
	// At the end of the method :
	// the total pair (pair + triplet) and interpolated triplet data files
    // are written in the 2 directories : 
	// pair, tripInterp
// ********************************************************************
{
	// to count the number of points in pure pair production file
	Int_t point3[101]; 
	// to count the total number of points in interpolated triplet file
	Int_t point2[101]; 
	// to count the number of triplet points actually requiring a calculation of interpolation
	Int_t point[101]; 

	for(Int_t z = 0;z<=100;++z)
	{
		point[z] = 0;
	}
	
	for(Int_t z = 0;z<=100;++z)
	{
		point2[z] = 0;
	}
	for(Int_t z = 0;z<=100;++z)
	{
		point3[z] = 0;
	}
	

	
	for(Int_t z = 1;z<=100;++z)  // loop for all the elements from 1 to 100
	{	// the name of file for pure pair production that will be read
		char pairFileFormat[64] = "epics2017/pairdata/pp-pair-cs-%d.dat";   
		// the name of file for triplet production that will be read
		char tripletFileFormat[64] = "epics2017/tripdata/pp-trip-cs-%d.dat"; 
		// the name of file for total pair production that will be written
		char ppFileFormat[64] = "epics2017/pair/pp-cs-%d.dat"; 
		// the name of file for interpolated triplet that will be written		
		char triplet_interpolatedFormat[64]= "epics2017/tripInterp/pp-tripInterp-cs-%d.dat"; 
		
		
		char pairFile[128] = "";   
		char tripletFile[128] = ""; 
		char ppFile[128] = ""; 
		char triplet_interpolated[128] = "";
		
		//appends z to the file name 
		sprintf(pairFile,pairFileFormat,z); 
		sprintf(tripletFile,tripletFileFormat,z);	
		sprintf(ppFile,ppFileFormat,z);	
		sprintf(triplet_interpolated,triplet_interpolatedFormat,z);	
		
		
		Int_t loop = 0;
		char str[64] = "";
		//save the number of energy points for pair production
		Int_t num1 = 0;  
		//save the number of energy points for triplet production
		Int_t num2 = 0; 
		Int_t ncols;
		double e, cs;
		
		
		cout<<"**********************************************************z==="<<z<<endl;
		//read pure pair production energy-cross section data and save them in x_1[num1] double y_1[num1];
		FILE * fp1 = fopen(pairFile,"r");
// ********************************************************************
// Read the two header lines to get the number of points
// of pure pair production data file (fp1 file)
// (this number is read at the beginning of second header line)
// ********************************************************************		
		while(1)  
		{	
			loop++;
			// if loop = 1 we read the first header line
			fgets (str, 64, fp1);
			// cout<<str<<endl;
			if(loop ==2)
			{
				// now loop = 2, we read the second header line, namely the number of points
				num1 = stoi(str);  
				cout<<"number of pure pair production points:"<<num1<<endl;
				break;
			}		
			
		}
		point3[z] = num1; 
		// array for pure pair production data: energy
		double x_1[num1];
		// array for pure pair production data: cross section		
		double y_1[num1]; 
		
		for(Int_t i = 0;i<num1;++i)
		{
			// read one line energy cross section
			ncols = fscanf(fp1,"%lf %lf",&e,&cs); 
			x_1[i]=e;   // keep energy
			y_1[i]=cs;  // keep cross section
			// printf("%.9e %.9e\n",x_1[i],y_1[i]);
		}
		fclose(fp1);
// ********************************************************************
// Read triplet production energy-cross section data and save them 
// in x_2[num2] double y_2[num2];
// ********************************************************************		
		
		
		FILE * fp2 = fopen(tripletFile,"r");		
		loop =0;
		
		while(1)
		{	
			loop++;
			// if loop = 1 we read the first header line
			fgets (str, 64, fp2);
			if(loop ==2)
			{
				// now loop = 2, we read the second header line, namely the number of points
				num2 = stoi(str);  
				cout<<"number of trip production points:"<<num2<<endl;
				break;
			}		
			
		}
		// array for triplet production data: energy
		double x_2[num2]; 
		// array for triplet production data: cross section
		double y_2[num2]; 
		
		for(Int_t i = 0;i<num2;++i)
		{
			// read one line energy cross section
			ncols = fscanf(fp2,"%lf %lf",&e,&cs); 
			x_2[i]=e; // keep energy
			y_2[i]=cs; // keep cross section
			//printf("%.9e %.9e\n",x_2[i],y_2[i]);
		}
		fclose(fp2);
		
		
		// array for total pair production data: energy
		double x_3[num1]; 
		// array for total pair production data: cross section
		double y_3[num1]; 
		
		// array for triplet_interpolated data: energy
		double x_4[num2];
		// array for triplet_interpolated data: cross section
		double y_4[num2];
		Int_t num4 = 0;
		
// ***************************************************************************	
// Create the total (pair + interpolated triplet) energy - cross section file
// ***************************************************************************
		
		// the number of total pp energy-cross section points will be 
		// the same to that of pure pair production, namely num1 
		
		for(Int_t i = 0; i<num1; ++i)
		{
			x_3[i] = x_1[i];  // the energy value for final total production data file is 
							  //the same as for pure pair production data file (taken as a reference)
							  
			y_3[i] = y_1[i];  // temporary value is pure pair production cs
			
			if(x_1[i]<x_2[0]) continue;  // if energy is lower than 2.044 MeV (first energy value for triplet)

			for(Int_t j =0; j<num2;++j)  // energy is above 2.044 MeV
			{
				if(IsEqual(x_2[j], x_1[i], eps, releps))     // the energy value in triplet file is the same as in pure pair 
				{
					y_3[i] = y_3[i] + y_2[j];  // total cs = pure pair cs + triplet cs
					// cout<<i<<"$$$"<<x_3[i]<<"-----"<<y_3[i]<<endl;
					//printf("No   %.9e pair: %.9e trip: %.9e %.9e\n",x_1[i],y_1[i],y_2[j],y_3[i]);
				
					x_4[num4] = x_2[j];  // the final interpolated triplet file is written in x_4 and y_4
					y_4[num4] = y_2[j];
					num4++;              // number of points for interpolated triplet file
					
					point2[z]++;					
					break;	             // (go to line 366)
				}	
				else if(x_2[j] > x_1[i]) 	// we are looking for the energy in triplet file
											// value just above x_1[i] (pure pair)
											
											// then we need to interpolate the triplet cs
											// between this point and the previous one
				{
					// two energy-cross section points for the interpolation
					
					double energy1, energy2;
					double cs1, cs2;
					
					energy1 = x_2[j-1];     // lowest energy point 
					cs1 = y_2[j-1];
					
					energy2 = x_2[j];       // highest energy point 
					cs2 = y_2[j];
					
					double cs_triplet_interpolated ;
					  
					
					// We use the linear interpolation	
					cs_triplet_interpolated = (cs2 - cs1)*(x_1[i] - energy1)/(energy2 - energy1) + cs1;
					// cout<<"start--------------------------------------------------------------"<<endl;
					// printf("log==%.9e\n", y);
					//cout<<"log=="<<y<<endl;
					// printf("x_1==%.9e\n",x_1[i]);
					
					// printf("--------(%.9e, %.9e),(%.9e, %.9e), (%.9e, %.9e)\n",x_1[i], cs_triplet_interpolated, energy1,cs1, energy2,cs2);
										
					y_3[i] = y_3[i] + cs_triplet_interpolated; //total (pure pair + interpolated triplet) cross section
					
					x_4[num4] = x_1[i];  // energy for interpolated triplet file
					y_4[num4] = cs_triplet_interpolated; // cs for interpolated triplet file
					num4++;     // number of points for interpolated triplet file
					
					point[z]++;
					point2[z]++;
					
					//printf("With %.9e pair: %.9e trip: %.9e %.9e\n",x_1[i],y_1[i],cs_triplet_interpolated,y_3[i]);
					// printf("%d*********x_3==%.9e eeeer-cssssss y_3==%.9e\n",i,x_3[i],y_3[i]);
					// cout<<"end---------------------------------------------------------------"<<endl;
					break;
				}
										
				
			}
			
			
		}
		//write total pp energy-cross section data: x_3[num1] double y_3[num1] in ppFile;
		FILE * fp3 = fopen(ppFile,"w");
		string input = "";  // save energy-cross section table
		double startEnergy,endEnergy;  // keep the minimum and maximum energy of the points for one element
		char headers [100] = ""; // save two headers
		
		for(Int_t i=0; i<num1;++i)
		{
			if(i == 0) startEnergy = x_3[i];
			if(i == num1-1) endEnergy = x_3[i]; 
			char energy_cs [64] = ""; // temporary variable, saving energy-cross section point
			sprintf(energy_cs,"%.9e %.9e\n", x_3[i], y_3[i]);
			input = input + energy_cs;
			
		}
		sprintf(headers,"%.9e %.9e %d\n%d\n", startEnergy, endEnergy, num1,num1); //save the two header lines in the variable headers
					
		input = headers + input; //concatenate two header lines with data table
		fprintf(fp3,"%s",input.c_str());
		fclose(fp3);		
		
		//write interpolated triplet data: x_4[num4] double y_4[num4] in triplet_interpolated;
		FILE * fp4 = fopen(triplet_interpolated,"w");	
		input = "";	
		cout<<"number of interpolated triplet: "<<num4<<endl;
		for(Int_t i=0; i<num4;++i)
		{
			if(i == 0) startEnergy = x_4[i];
			if(i == num4-1) endEnergy = x_4[i];
			char energy_cs [64] = ""; // temporary variable, saving energy-cross section point
			sprintf(energy_cs,"%.9e %.9e\n", x_4[i], y_4[i]);
			input = input + energy_cs; //concatenate two header lines with data table
		}
		
		sprintf(headers,"%.9e %.9e %d\n%d\n", startEnergy, endEnergy, num4,num4);
		input = headers + input;
		fprintf(fp4,"%s",input.c_str());
		fclose(fp4);		
				
		
	}
	
	FILE * fp5 = fopen("out.txt","w");	
	for(Int_t z=1; z<=100;++z)
	{
		fprintf(fp5,"Element = %3d: Number of points in pure pair production file = %d \n \
		       Total triplet points = %d; Triplet requiring interpolation calculation = %d \n", z,point3[z], point2[z], point[z]);
	}

	fclose(fp5);	
	
}

// main method

void generate_pp_cs()
{
	string dirs[5] = {"epics2017","pairdata","tripdata", "pair", "tripInterp"};
	// create the 5 directories if they do not exist
	for(Int_t i =0;i<5;++i)
	{
		if(i==0)
		{
			dirs[i] = "mkdir -p "+dirs[i]; 
			// execute the command defined line above
			gSystem->Exec(dirs[i].c_str());
		}
		else 
		{
			dirs[i] = "mkdir -p epics2017/" + dirs[i];
			// execute the command defined line above			
			gSystem->Exec(dirs[i].c_str());
		}
	}
	
	extract_pair_trip_cs(); 
	// now the pure pair and triplet data files are written in the 2 directories : 
	// pairdata, tripdata 
	
	// we interpolate triplet data so that we get the same energy points 
	// for pair (taken as a reference) and triplet 
	create_pp_cs();	
	// now the total pair (pair + triplet) and interpolated triplet data files
    // are written in the 2 directories : 
	// pair, tripInterp
}
